'''
Initial cut at a centralized file for the IO
of the large single aisle 1 V3 bug fixed case for the GASP-based
mass calculations and geometry calculations.
'''
import numpy as np

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission

V3_bug_fixed_options = get_option_defaults()
V3_bug_fixed_options.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM,
                             val=False, units='unitless')
V3_bug_fixed_options.set_val(
    Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
V3_bug_fixed_options.set_val(Mission.Design.CRUISE_ALTITUDE, val=37500, units='ft')
V3_bug_fixed_options.set_val(
    Aircraft.Wing.CHOOSE_FOLD_LOCATION, val=False, units='unitless')
V3_bug_fixed_options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                             val=True, units='unitless')
V3_bug_fixed_options.set_val(
    Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, val=True, units='unitless')
V3_bug_fixed_options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                             val=False, units='unitless')


V3_bug_fixed_options.set_val(
    Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
)
V3_bug_fixed_options.set_val(Aircraft.Wing.SWEEP, val=25, units="deg")
V3_bug_fixed_options.set_val(
    Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
)
V3_bug_fixed_options.set_val(
    Mission.Design.GROSS_MASS, val=175400, units="lbm"
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.LOADING, val=126, units="lbf/ft**2"
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.MOMENT_RATIO, val=0.2307, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.REFERENCE_SLS_THRUST, np.array([28690]), units="lbf"
)
# NOTE override required for mass summation test
V3_bug_fixed_options.set_val(
    Aircraft.Engine.SCALE_FACTOR, 1.02823, units='unitless'
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.FUEL_MARGIN, val=0, units="unitless")


V3_bug_fixed_non_metadata = AviaryValues()


V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.NUM_SEATS_ABREAST, 6, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.SEAT_WIDTH, 20.2, units="inch"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.NUM_AISLES, 1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.AISLE_WIDTH, 24, units="inch"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.SEAT_PITCH, 29, units="inch"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.DELTA_DIAMETER, 4.5, units="ft"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, 9.5, units="ft")
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.NOSE_FINENESS, 1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.TAIL_FINENESS, 3, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.WETTED_AREA_FACTOR, 4000, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.MOMENT_RATIO, 2.362, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless"
)
V3_bug_fixed_options.set_val(Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
V3_bug_fixed_options.set_val(
    Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless")
V3_bug_fixed_options.set_val(Aircraft.Nacelle.FINENESS, 2, units="unitless")

V3_bug_fixed_options.set_val(
    Aircraft.Design.MAX_STRUCTURAL_SPEED, val=402.5, units="mi/h"
)
V3_bug_fixed_options.set_val(
    Aircraft.Design.LIFT_CURVE_SLOPE, val=7.1765, units="1/rad"
)
V3_bug_fixed_options.set_val(
    Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS, val=200, units="lbm"
)
V3_bug_fixed_options.set_val(
    Aircraft.CrewPayload.CARGO_MASS, val=10040, units="lbm"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.SWEEP, val=0, units='deg'
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.MASS_COEFFICIENT, val=0.232, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.LandingGear.TAIL_HOOK_MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.MASS_COEFFICIENT, val=0.289, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
)
V3_bug_fixed_options.set_val(
    # TODO: should be lbf/ft**2 I think
    Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT, val=2.66, units='unitless'
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.SURFACE_CONTROL_MASS_COEFFICIENT, val=0.95, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Design.COCKPIT_CONTROL_MASS_COEFFICIENT, val=16.5, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS, val=0, units="lbm"
)
V3_bug_fixed_options.set_val(
    Aircraft.Controls.COCKPIT_CONTROL_MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.Controls.CONTROL_MASS_INCREMENT, val=0, units="lbm"
)
V3_bug_fixed_options.set_val(
    Aircraft.LandingGear.MASS_COEFFICIENT, val=0.04, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT, val=0.85, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Nacelle.CLEARANCE_RATIO, val=0.2, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.MASS_SPECIFIC, val=0.21366, units="lbm/lbf"
)
V3_bug_fixed_options.set_val(
    Aircraft.Nacelle.MASS_SPECIFIC, val=3, units="lbm/ft**2"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.PYLON_FACTOR, val=1.25, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.ADDITIONAL_MASS_FRACTION, val=0.14, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.Propulsion.MISC_MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0.15, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
    val=[
        928.0,
        0.0736,
        0.112,
        0.14,
        1959.0,
        1.65,
        551.0,
        11192.0,
        5.0,
        3.0,
        50.0,
        7.6,
        12.0,
    ],
    units="unitless",
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.MASS_COEFFICIENT, val=102.5, units="unitless"
)

V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.MASS_COEFFICIENT, val=128, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.HorizontalTail.MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.VerticalTail.MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuselage.MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Engine.POD_MASS_SCALER, val=1, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Design.STRUCTURAL_MASS_INCREMENT, val=0, units='lbm'
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS"
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT, val=1.9, units="unitless"
)  # Based on large single aisle 1 for updated flaps mass model
V3_bug_fixed_options.set_val(
    Mission.Landing.LIFT_COEFFICIENT_MAX, val=2.817, units="unitless"
)  # Based on large single aisle 1 for updated flaps mass model
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
V3_bug_fixed_options.set_val(
    Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless"
)
V3_bug_fixed_options.set_val(
    Aircraft.Wing.FLAP_CHORD_RATIO, val=0.3, units='unitless'
)


V3_bug_fixed_non_metadata.set_val(
    "fuel_mass.fus_and_struct.pylon_len", val=0, units='ft'
)
V3_bug_fixed_non_metadata.set_val(
    "fuel_mass.fus_and_struct.MAT", val=0, units='lbm'
)



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.design import Design
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Mission


class DesignMCLTest(unittest.TestCase):

    def test_derivs_supersonic(self):
        prob = om.Problem()
        model = prob.model

        options = {}
        options[Aircraft.Wing.AIRFOIL_TECHNOLOGY] = (1.0, 'unitless')
        options[Mission.Constraints.MAX_MACH] = (0.9, 'unitless')

        model.add_subsystem(
            'design', Design(aviary_options=AviaryValues(options)), promotes_inputs=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05)
        prob.set_val(Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, val=1.0)
        prob.set_val(Aircraft.Wing.SWEEP, val=2.0191)
        prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD, val=0.12)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)

    def test_derivs_subsonic(self):
        prob = om.Problem()
        model = prob.model

        options = {}
        options[Aircraft.Wing.AIRFOIL_TECHNOLOGY] = (1.0, 'unitless')
        options[Mission.Constraints.MAX_MACH] = (0.9, 'unitless')

        model.add_subsystem(
            'design', Design(aviary_options=AviaryValues(options)), promotes_inputs=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05)
        prob.set_val(Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, val=1.0)
        prob.set_val(Aircraft.Wing.SWEEP, val=2.191)
        prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD, val=0.12)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.constants import RHO_SEA_LEVEL_ENGLISH

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic

from .unsteady_solved_eom import UnsteadySolvedEOM


class UnsteadyControlIterGroup(om.Group):
    """
    This Group contains a nonlinear solver to determine alpha and thrust for a given flight condition.
    """

    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare("ground_roll", types=bool, default=False,
                             desc="True if the aircraft is confined to the ground. Removes altitude rate as an "
                                  "output and adjusts the TAS rate equation.")
        self.options.declare("clean", types=bool, default=False,
                             desc="If true then no flaps or gear are included. Useful for high-speed flight phases.")
        self.options.declare(
            'aviary_options', types=AviaryValues, default=None,
            desc='collection of Aircraft/Mission specific options'
        )

        # TODO finish description
        self.options.declare(
            'core_subsystems',
            desc='list of core subsystems'
        )

        self.options.declare(
            'subsystem_options', types=dict, default={},
            desc='dictionary of parameters to be passed to the subsystem builders'
        )

    def setup(self):
        nn = self.options["num_nodes"]
        ground_roll = self.options["ground_roll"]
        clean = self.options["clean"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options}

        if clean:
            kwargs['method'] = 'low_speed'
        else:
            kwargs['method'] = 'cruise'

        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                self.add_subsystem(subsystem.name,
                                   system,
                                   promotes_inputs=subsystem.mission_inputs(**kwargs),
                                   promotes_outputs=subsystem.mission_outputs(**kwargs))

        eom_comp = UnsteadySolvedEOM(num_nodes=nn, ground_roll=ground_roll)

        self.add_subsystem("eom", subsys=eom_comp,
                           promotes_inputs=["*",
                                            (Dynamic.Mission.THRUST_TOTAL, "thrust_req")],
                           promotes_outputs=["*"])

        thrust_alpha_bal = om.BalanceComp()
        if not self.options['ground_roll']:
            thrust_alpha_bal.add_balance("alpha",
                                         units="rad",
                                         val=np.zeros(nn),
                                         lhs_name="dgam_dt_approx",
                                         rhs_name="dgam_dt",
                                         eq_units="rad/s",
                                         normalize=False)

        thrust_alpha_bal.add_balance("thrust_req",
                                     units="N",
                                     val=100*np.ones(nn),
                                     lhs_name="dTAS_dt_approx",
                                     rhs_name="dTAS_dt",
                                     eq_units="m/s**2",
                                     normalize=False)

        self.add_subsystem("thrust_alpha_bal", subsys=thrust_alpha_bal,
                           promotes_inputs=["*"],
                           promotes_outputs=["*"])

        self.nonlinear_solver = om.NewtonSolver(solve_subsystems=True,
                                                atol=1.0e-10,
                                                rtol=1.0e-10)
        # self.nonlinear_solver.linesearch = om.ArmijoGoldsteinLS()
        self.linear_solver = om.DirectSolver(assemble_jac=True)

        # Set common default values for promoted inputs
        onn = np.ones(nn)
        self.set_input_defaults(
            name="rho", val=RHO_SEA_LEVEL_ENGLISH * onn, units="slug/ft**3")
        self.set_input_defaults(
            name=Dynamic.Mission.SPEED_OF_SOUND,
            val=1116.4 * onn,
            units="ft/s")
        if not self.options['ground_roll']:
            self.set_input_defaults(name=Dynamic.Mission.FLIGHT_PATH_ANGLE,
                                    val=0.0 * onn, units="rad")
        self.set_input_defaults(name="TAS", val=250. * onn, units="kn")



import unittest
from pathlib import Path

from aviary.utils.functions import get_path
from openmdao.utils.testing_utils import use_tempdirs

from aviary.utils.Fortran_to_Aviary import LegacyCode, _exec_F2A


class DummyArgs(object):
    def __init__(self):
        self.input_deck = None
        self.out_file = None
        self.legacy_code = None
        self.defaults_deck = False
        self.force = False


@use_tempdirs
class TestFortranToAviary(unittest.TestCase):
    def prepare_and_run(self, filepath, out_file=None, legacy_code=LegacyCode.GASP):
        args = DummyArgs()

        # Specify the input file and the legacy code
        args.input_deck = filepath

        # Specify the output file
        filename = filepath.split('.')[0]+'.csv'
        if not out_file:
            args.out_file = Path.cwd() / Path('TEST_'+filename)
        else:
            args.out_file = Path(out_file)
        args.legacy_code = legacy_code

        # Execute the conversion
        _exec_F2A(args, None)

    def compare_files(self, filepath, skip_list=[]):
        """
        Compares the converted file with a validation file.

        Use the `skip_list` input to specify strings that are in lines you want
        to skip. This is useful for skipping data that Aviary might need but
        Fortran-based tools do not.
        """
        filename = filepath.split('.')[0]+'.csv'

        validation_data = get_path(filename)

        # Open the converted and validation files
        with open('TEST_'+filename, 'r') as f_in, open(validation_data, 'r') as expected:
            for line in f_in:
                for skip in skip_list:
                    if skip in line:
                        break
                # Remove whitespace and compare
                expected_line = ''.join(expected.readline().split())
                line_no_whitespace = ''.join(line.split())

                # Assert that the lines are equal
                try:
                    self.assertEqual(line_no_whitespace.count(expected_line), 1)

                except Exception as error:
                    exc_string = f'Error:  {filename}\nFound: {line_no_whitespace}\nExpected:  {expected_line}'
                    raise Exception(exc_string)

    def test_large_single_aisle(self):
        filepath = 'models/large_single_aisle_1/large_single_aisle_1_GwGm.dat'

        self.prepare_and_run(filepath)
        self.compare_files(filepath)

    def test_small_single_aisle(self):
        filepath = 'models/small_single_aisle/small_single_aisle_GwGm.dat'

        self.prepare_and_run(filepath)
        self.compare_files(filepath)

    def test_diff_configuration(self):
        filepath = 'models/test_aircraft/converter_configuration_test_data_GwGm.dat'

        self.prepare_and_run(filepath)
        self.compare_files(filepath)

    def test_N3CC(self):
        # Note: The csv comparison file N3CC_generic_low_speed_polars_FLOPSinp.csv was generated using the fortran-to-Aviary converter
        # and was not evaluated for comparison to the original. Thus, until this file is evaluated, this test is purely a regression
        # test.

        filepath = 'models/N3CC/N3CC_generic_low_speed_polars_FLOPSinp.txt'
        out_file = Path.cwd() / 'TEST_models/N3CC/N3CC_generic_low_speed_polars_FLOPSinp.csv'

        self.prepare_and_run(filepath, out_file=out_file, legacy_code=LegacyCode.FLOPS)
        self.compare_files(
            'models/N3CC/N3CC_generic_low_speed_polars_FLOPSinp.csv')


if __name__ == "__main__":
    unittest.main()



from aviary.variable_info.variables import Aircraft


def mass_defaults(prob):
    prob.model.set_input_defaults(
        Aircraft.Strut.ATTACHMENT_LOCATION, val=0, units='unitless')
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless")
    prob.model.set_input_defaults(
        Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless")
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
    )

    prob.model.set_input_defaults(
        Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuselage.DELTA_DIAMETER, 4.5, units="ft"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, 9.5, units="ft"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuselage.NOSE_FINENESS, 1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuselage.TAIL_FINENESS, 3, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuselage.WETTED_AREA_FACTOR, 4000, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.MOMENT_RATIO, 2.362, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless"
    )
    prob.model.set_input_defaults(Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
    prob.model.set_input_defaults(
        Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Nacelle.FINENESS, 2, units="unitless")

    prob.model.set_input_defaults(
        Aircraft.Design.MAX_STRUCTURAL_SPEED,
        val=402.5,
        units="mi/h",
    )
    prob.model.set_input_defaults(
        Aircraft.Design.LIFT_CURVE_SLOPE,
        val=7.1765,
        units="unitless",
    )

    prob.model.set_input_defaults(
        Aircraft.CrewPayload.CARGO_MASS, val=10040, units="lbm"
    )
    prob.model.set_input_defaults(
        Aircraft.HorizontalTail.MASS_COEFFICIENT,
        val=0.232,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.LandingGear.TAIL_HOOK_MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.MASS_COEFFICIENT,
        val=0.289,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
        val=0.12,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.THICKNESS_TO_CHORD,
        val=0.12,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT,
        val=2.66,
        units="lbm/ft**2",
    )
    prob.model.set_input_defaults(
        Aircraft.Wing.SURFACE_CONTROL_MASS_COEFFICIENT,
        val=0.95,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.Design.COCKPIT_CONTROL_MASS_COEFFICIENT,
        val=16.5,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS, val=0, units="lbm"
    )
    prob.model.set_input_defaults(
        Aircraft.Controls.COCKPIT_CONTROL_MASS_SCALER, val=1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER, val=1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Controls.TOTAL_MASS,
        val=0,
        units="lbm",
    )
    prob.model.set_input_defaults(
        Aircraft.LandingGear.MASS_COEFFICIENT, val=0.04,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT, val=0.85,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Nacelle.CLEARANCE_RATIO,
        val=0.2,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.Engine.MASS_SPECIFIC,
        val=0.21366,
        units="lbm/lbf",
    )
    prob.model.set_input_defaults(
        Aircraft.Nacelle.MASS_SPECIFIC,
        val=3,
        units="lbm/ft**2",
    )
    prob.model.set_input_defaults(
        Aircraft.Engine.PYLON_FACTOR, val=1.25,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Engine.ADDITIONAL_MASS_FRACTION, val=0.14,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Engine.MASS_SCALER, val=1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Propulsion.MISC_MASS_SCALER, val=1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.LandingGear.MAIN_GEAR_LOCATION,
        val=0.15,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
        val=[
            928.0,
            0.0736,
            0.112,
            0.14,
            1959.0,
            1.65,
            551.0,
            11192.0,
            5.0,
            3.0,
            50.0,
            7.6,
            12.0,
        ],
        units="unitless",
    )

    prob.model.set_input_defaults(
        Aircraft.Wing.MASS_COEFFICIENT,
        val=102.5,
        units="unitless",
    )

    prob.model.set_input_defaults(
        Aircraft.Fuselage.MASS_COEFFICIENT,
        val=128,
        units="unitless",
    )
    prob.model.set_input_defaults(
        "static_analysis.total_mass.fuel_mass.fus_and_struct.pylon_len",
        val=0,
        units='ft',
    )
    prob.model.set_input_defaults(
        "static_analysis.total_mass.fuel_mass.fus_and_struct.MAT", val=0,
        units='lbm'
    )
    prob.model.set_input_defaults(
        Aircraft.Wing.MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.HorizontalTail.MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuselage.MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Engine.POD_MASS_SCALER, val=1,
        units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Design.STRUCTURAL_MASS_INCREMENT,
        val=0,
        units='lbm',
    )
    prob.model.set_input_defaults(
        Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless"
    )
    prob.model.set_input_defaults(
        Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT,
        val=0.041,
        units="unitless",
    )
    prob.model.set_input_defaults(
        Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS"
    )
    prob.model.set_input_defaults(
        Aircraft.VerticalTail.SWEEP,
        val=25.0,
        units="deg",
    )

    return prob



from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime, InitialGuessControl
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from aviary.mission.gasp_based.ode.accel_ode import AccelODE


class AccelPhase(PhaseBuilderBase):
    """
    A phase builder for an acceleration phase in a mission simulation.

    This class extends the PhaseBuilderBase class, providing specific implementations for
    the acceleration phase of a flight mission.

    Attributes
    ----------
    Inherits all attributes from PhaseBuilderBase.

    Methods
    -------
    Inherits all methods from PhaseBuilderBase.
    Additional method overrides and new methods specific to the acceleration phase are included.
    """
    default_name = 'accel_phase'
    default_ode_class = AccelODE

    _meta_data_ = {}
    _initial_guesses_meta_data_ = {}

    def build_phase(self, aviary_options: AviaryValues = None):
        """
        Return a new acceleration phase for analysis using these constraints.

        Parameters
        ----------
        aviary_options : AviaryValues
            Collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        """
        phase = self.phase = super().build_phase(aviary_options)
        user_options = self.user_options

        # Extracting and setting options
        EAS_constraint_eq = user_options.get_val('EAS_constraint_eq', units='kn')
        alt = user_options.get_val('alt', units='ft')

        # States
        self.add_velocity_state(user_options)

        self.add_mass_state(user_options)

        self.add_distance_state(user_options)

        # Boundary Constraints
        phase.add_boundary_constraint(
            "EAS", loc="final", equals=EAS_constraint_eq, units="kn", ref=EAS_constraint_eq
        )

        phase.add_parameter(Dynamic.Mission.ALTITUDE, opt=False, units="ft", val=alt)

        # Timeseries Outputs
        phase.add_timeseries_output("EAS", output_name="EAS", units="kn")
        phase.add_timeseries_output(
            Dynamic.Mission.MACH, output_name=Dynamic.Mission.MACH, units="unitless")
        phase.add_timeseries_output("alpha", output_name="alpha", units="deg")
        phase.add_timeseries_output("aero.CL", output_name="CL", units="unitless")
        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL, output_name=Dynamic.Mission.THRUST_TOTAL, units="lbf")
        phase.add_timeseries_output("aero.CD", output_name="CD", units="unitless")

        return phase


# Adding metadata for the AccelPhase
AccelPhase._add_meta_data('fix_initial', val=False)
AccelPhase._add_meta_data('EAS_constraint_eq', val=250, units='kn')
AccelPhase._add_meta_data('duration_bounds', val=(0, 0), units='s')
AccelPhase._add_meta_data('duration_ref', val=1, units='s')
AccelPhase._add_meta_data('velocity_lower', val=0, units='kn')
AccelPhase._add_meta_data('velocity_upper', val=0, units='kn')
AccelPhase._add_meta_data('velocity_ref', val=1, units='kn')
AccelPhase._add_meta_data('velocity_ref0', val=0, units='kn')
AccelPhase._add_meta_data('velocity_defect_ref', val=None, units='kn')
AccelPhase._add_meta_data('mass_lower', val=0, units='lbm')
AccelPhase._add_meta_data('mass_upper', val=0, units='lbm')
AccelPhase._add_meta_data('mass_ref', val=1, units='lbm')
AccelPhase._add_meta_data('mass_ref0', val=0, units='lbm')
AccelPhase._add_meta_data('mass_defect_ref', val=None, units='lbm')
AccelPhase._add_meta_data('distance_lower', val=0, units='NM')
AccelPhase._add_meta_data('distance_upper', val=0, units='NM')
AccelPhase._add_meta_data('distance_ref', val=1, units='NM')
AccelPhase._add_meta_data('distance_ref0', val=0, units='NM')
AccelPhase._add_meta_data('distance_defect_ref', val=None, units='NM')
AccelPhase._add_meta_data('alt', val=500, units='ft')
AccelPhase._add_meta_data('num_segments', val=None, units='unitless')
AccelPhase._add_meta_data('order', val=None, units='unitless')

AccelPhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for initial time and duration specified as a tuple')

AccelPhase._add_initial_guess_meta_data(
    InitialGuessState('velocity'),
    desc='initial guess for true airspeed')

AccelPhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass')

AccelPhase._add_initial_guess_meta_data(
    InitialGuessState('distance'),
    desc='initial guess for horizontal distance traveled')

AccelPhase._add_initial_guess_meta_data(
    InitialGuessControl('throttle'),
    desc='initial guess for throttle')



import numpy as np
import openmdao.api as om

from pathlib import Path

from aviary.subsystems.aerodynamics.flops_based.drag import TotalDrag as Drag
from aviary.subsystems.aerodynamics.flops_based.lift import \
    LiftEqualsWeight as CL
from aviary.utils.csv_data_file import read_data_file
from aviary.utils.functions import get_path
from aviary.utils.named_values import NamedValues
from aviary.variable_info.variables import Aircraft, Dynamic
from aviary.utils.data_interpolator_builder import build_data_interpolator


# Map of variable names to allowed headers for data files (only lowercase required,
# spaces are replaced with underscores when data tables are read)
# "Repeated" aliases allows variables with different cases to match with desired
# all-lowercase name
aliases = {Dynamic.Mission.ALTITUDE: ['h', 'alt', 'altitude'],
           Dynamic.Mission.MACH: ['m', 'mach'],
           'lift_coefficient': ['cl', 'coefficient_of_lift', 'lift_coefficient'],
           'lift_dependent_drag_coefficient': ['cdi', 'lift_dependent_drag_coefficient',
                                               'lift-dependent_drag_coefficient'],
           'zero_lift_drag_coefficient': ['cd0', 'zero_lift_drag_coefficient',
                                          'zero-lift_drag_coefficient'],
           }


class TabularAeroGroup(om.Group):
    def initialize(self):
        options = self.options

        options.declare('num_nodes', types=int)

        options.declare('CD0_data', types=(str, Path, NamedValues),
                        desc='Data file or NamedValues object containing zero-lift drag '
                             'coefficient table.')

        options.declare('CDI_data', types=(str, Path, NamedValues),
                        desc='Data file or NamedValues object containing lift-dependent '
                             'drag coefficient table.')

        options.declare('structured', types=bool, default=True,
                        desc='Flag that sets if data is a structured grid.')

        options.declare('training_data', types=bool, default=False,
                        desc='Flag that sets if drag data for interpolation will be '
                             'passed via openMDAO connections. If True, provided values  '
                             'for drag coefficients in data will be ignored.')

    def setup(self):
        options = self.options
        nn = options['num_nodes']
        CDI_table = options['CDI_data']
        CD0_table = options['CD0_data']
        structured = options['structured']
        training_data = options['training_data']

        # if data is from file, read data using alias dict
        if isinstance(CDI_table, str):
            CDI_table = get_path(CDI_table)
        if isinstance(CDI_table, Path):
            CDI_table = read_data_file(CDI_table, aliases=aliases)
        if isinstance(CD0_table, str):
            CD0_table = get_path(CD0_table)
        if isinstance(CD0_table, Path):
            CD0_table = read_data_file(CD0_table, aliases=aliases)

        if training_data or not structured:
            method = 'lagrange3'
        else:
            method = '2D-lagrange3'

        CD0_interp = build_data_interpolator(nn, interpolator_data=CD0_table,
                                             interpolator_outputs={'zero_lift_drag_coefficient':
                                                                   'unitless'},
                                             method=method, structured=structured,
                                             training_data=training_data)

        CDI_interp = build_data_interpolator(nn, interpolator_data=CDI_table,
                                             interpolator_outputs={'lift_dependent_drag_coefficient':
                                                                   'unitless'},
                                             method=method, structured=structured,
                                             training_data=training_data)

        # add subsystems
        self.add_subsystem(
            Dynamic.Mission.DYNAMIC_PRESSURE, _DynamicPressure(num_nodes=nn),
            promotes_inputs=[Dynamic.Mission.VELOCITY, Dynamic.Mission.DENSITY],
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE])

        self.add_subsystem(
            'lift_coefficient', CL(num_nodes=nn),
            promotes_inputs=[Dynamic.Mission.MASS,
                             Aircraft.Wing.AREA, Dynamic.Mission.DYNAMIC_PRESSURE],
            promotes_outputs=[('cl', 'lift_coefficient'), Dynamic.Mission.LIFT])

        self.add_subsystem('CD0_interp', CD0_interp,
                           promotes_inputs=['*'],
                           promotes_outputs=['*'])

        self.add_subsystem('CDI_interp', CDI_interp,
                           promotes_inputs=['*'],
                           promotes_outputs=['*'])

        self.add_subsystem(
            Dynamic.Mission.DRAG, Drag(num_nodes=nn),
            promotes_inputs=[
                Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR,
                Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR,
                Aircraft.Wing.AREA,
                Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR,
                Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR,
                ('CDI', 'lift_dependent_drag_coefficient'), ('CD0', 'zero_lift_drag_coefficient'), Dynamic.Mission.MACH, Dynamic.Mission.DYNAMIC_PRESSURE],
            promotes_outputs=['drag_coefficient', Dynamic.Mission.DRAG])


class _DynamicPressure(om.ExplicitComponent):
    '''
    Calculate dynamic pressure as a function of velocity and density.
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(Dynamic.Mission.VELOCITY, val=np.ones(nn), units='m/s')
        self.add_input(Dynamic.Mission.DENSITY, val=np.ones(nn), units='kg/m**3')

        self.add_output(
            Dynamic.Mission.DYNAMIC_PRESSURE, val=np.ones(nn), units='N/m**2',
            desc='pressure caused by fluid motion')

    def setup_partials(self):
        nn = self.options['num_nodes']

        rows_cols = np.arange(nn)

        self.declare_partials(
            Dynamic.Mission.DYNAMIC_PRESSURE, [
                Dynamic.Mission.VELOCITY, Dynamic.Mission.DENSITY],
            rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs):
        TAS = inputs[Dynamic.Mission.VELOCITY]
        rho = inputs[Dynamic.Mission.DENSITY]

        outputs[Dynamic.Mission.DYNAMIC_PRESSURE] = 0.5 * rho * TAS**2

    def compute_partials(self, inputs, partials):
        TAS = inputs[Dynamic.Mission.VELOCITY]
        rho = inputs[Dynamic.Mission.DENSITY]

        partials[Dynamic.Mission.DYNAMIC_PRESSURE, Dynamic.Mission.VELOCITY] = rho * TAS
        partials[Dynamic.Mission.DYNAMIC_PRESSURE,
                 Dynamic.Mission.DENSITY] = 0.5 * TAS**2



from aviary.utils.aviary_values import AviaryValues
from aviary.utils.preprocessors import preprocess_propulsion
from aviary.utils.functions import get_path
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.variables import Aircraft
from aviary.subsystems.propulsion.engine_deck import EngineDeck


def get_option_defaults(engine=True, meta_data=_MetaData) -> AviaryValues:
    """
    Returns a deep copy of the collection of all options for which default values exist.

    Parameters
    ----------
    engine : bool
        If true, the collection includes the default engine model
    meta_data : dict
        Dictionary containing metadata for the options. If None, Aviary's built-in
        metadata will be used.
    """

    option_defaults = AviaryValues()

    # Load all variables marked as options in the MetaData
    for key in meta_data:
        var = meta_data[key]
        if var['option'] and var['default_value'] is not None:
            option_defaults.set_val(key, var['default_value'], var['units'])

    if engine:
        engine_options = option_defaults.deepcopy()
        engine_options.set_val(Aircraft.Engine.DATA_FILE,
                               get_path('models/engines/turbofan_23k_1.deck'))
        engine_options.set_val(Aircraft.Engine.SCALE_FACTOR,
                               meta_data[Aircraft.Engine.SCALE_FACTOR]['default_value'])
        engine_deck = EngineDeck(options=engine_options)
        preprocess_propulsion(option_defaults, [engine_deck])

    return option_defaults


def is_option(key, meta_data=_MetaData) -> bool:
    """
    Returns True if the variable is defined as an option in the MetaData.

    Parameters
    ----------
    key: str
        Name of the variable to be checked
    meta_data : dict
        Dictionary containing metadata for the variable. If None, Aviary's built-in
        metadata will be used.
    """

    return meta_data[key]['option']



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_GASP, GRAV_ENGLISH_LBM, MU_LANDING
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class LandingAltitudeComponent(om.ExplicitComponent):
    def setup(self):
        add_aviary_input(self, Mission.Landing.OBSTACLE_HEIGHT, val=50.0)
        add_aviary_input(self, Mission.Landing.AIRPORT_ALTITUDE, val=0.0)

        add_aviary_output(self, Mission.Landing.INITIAL_ALTITUDE, val=0.0)

        self.declare_partials(Mission.Landing.INITIAL_ALTITUDE, "*", val=1)

    def compute(self, inputs, outputs):
        (
            approach_height,
            airport_alt,
        ) = inputs.values()
        outputs[Mission.Landing.INITIAL_ALTITUDE] = airport_alt + approach_height

    def compute_partials(self, inputs, J):
        pass


class GlideConditionComponent(om.ExplicitComponent):
    def setup(self):
        self.add_input("rho_app", val=0.0, units="slug/ft**3", desc="air density")
        add_aviary_input(self, Mission.Landing.MAXIMUM_SINK_RATE, val=900.0)
        self.add_input(Dynamic.Mission.MASS, val=0.0, units="lbm",
                       desc="aircraft mass at start of landing")
        add_aviary_input(self, Aircraft.Wing.AREA, val=1.0)
        add_aviary_input(self, Mission.Landing.GLIDE_TO_STALL_RATIO, val=1.3)
        self.add_input("CL_max", val=0.0, units='unitless',
                       desc="CLMX: max CL at approach altitude")

        add_aviary_input(self, Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR, val=1.15)
        add_aviary_input(self, Mission.Landing.TOUCHDOWN_SINK_RATE, val=5.0)
        add_aviary_input(self, Mission.Landing.INITIAL_ALTITUDE, val=0.0)
        add_aviary_input(self, Mission.Landing.BRAKING_DELAY, val=1.0)

        add_aviary_output(self, Mission.Landing.INITIAL_VELOCITY, val=0.0,
                          desc="glide speed calculated using TAS_stall")
        add_aviary_output(self, Mission.Landing.STALL_VELOCITY, val=0.0)

        self.add_output(
            "TAS_touchdown", val=0.0, units="ft/s", desc="VTD: touchdown speed"
        )
        self.add_output(
            "density_ratio", val=0.0, units="unitless", desc="DRAT: density ratio for DLAND"
        )
        self.add_output(
            "wing_loading_land",
            val=0.0,
            units="lbf/ft**2",
            desc="WOS: wing loading at landing",
        )
        self.add_output(
            "theta",
            val=0.0,
            units="rad",
            desc="THETA: theta angle for approach to transition",
        )

        self.add_output(
            "glide_distance", val=0.0, units="ft", desc="DLGL: glide distance"
        )
        self.add_output(
            "tr_distance",
            val=0.0,
            units="ft",
            desc="DLTR: distance covered by the flare maneuver",
        )
        self.add_output(
            "delay_distance",
            val=0.0,
            units="ft",
            desc="DDELAY: delay distance - touchdown to brake application",
        )
        self.add_output(
            "flare_alt", val=0.0, units="ft", desc="HFLAR: altitude of flare maneuver"
        )

        self.declare_partials(
            Mission.Landing.INITIAL_VELOCITY,
            [Dynamic.Mission.MASS, Aircraft.Wing.AREA, "CL_max", "rho_app",
                Mission.Landing.GLIDE_TO_STALL_RATIO],
        )
        self.declare_partials(
            Mission.Landing.STALL_VELOCITY, [
                Dynamic.Mission.MASS, Aircraft.Wing.AREA, "CL_max", "rho_app"]
        )
        self.declare_partials(
            "TAS_touchdown",
            [Mission.Landing.GLIDE_TO_STALL_RATIO, Dynamic.Mission.MASS,
                Aircraft.Wing.AREA, "CL_max", "rho_app"],
        )
        self.declare_partials("density_ratio", ["rho_app"])
        self.declare_partials("wing_loading_land", [
                              Dynamic.Mission.MASS, Aircraft.Wing.AREA])
        self.declare_partials(
            "theta",
            [
                Mission.Landing.MAXIMUM_SINK_RATE,
                Dynamic.Mission.MASS,
                Aircraft.Wing.AREA,
                "CL_max",
                "rho_app",
                Mission.Landing.GLIDE_TO_STALL_RATIO,
            ],
        )
        self.declare_partials(
            "glide_distance",
            [
                Mission.Landing.INITIAL_ALTITUDE,
                Mission.Landing.MAXIMUM_SINK_RATE,
                Dynamic.Mission.MASS,
                Aircraft.Wing.AREA,
                "CL_max",
                "rho_app",
                Mission.Landing.GLIDE_TO_STALL_RATIO,
            ],
        )
        self.declare_partials(
            "tr_distance",
            [
                Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR,
                Mission.Landing.TOUCHDOWN_SINK_RATE,
                Dynamic.Mission.MASS,
                Aircraft.Wing.AREA,
                "CL_max",
                "rho_app",
                Mission.Landing.GLIDE_TO_STALL_RATIO,
                Mission.Landing.MAXIMUM_SINK_RATE,
            ],
        )
        self.declare_partials(
            "delay_distance",
            [
                Mission.Landing.GLIDE_TO_STALL_RATIO,
                Dynamic.Mission.MASS,
                Aircraft.Wing.AREA,
                "CL_max",
                "rho_app",
                Mission.Landing.BRAKING_DELAY,
            ],
        )
        self.declare_partials(
            "flare_alt",
            [
                Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR,
                Mission.Landing.TOUCHDOWN_SINK_RATE,
                Mission.Landing.MAXIMUM_SINK_RATE,
                Dynamic.Mission.MASS,
                Aircraft.Wing.AREA,
                "CL_max",
                "rho_app",
                Mission.Landing.GLIDE_TO_STALL_RATIO,
            ],
        )

    def compute(self, inputs, outputs):
        (
            rho_app,
            rate_of_sink_max,
            mass,
            wing_area,
            glide_to_stall_ratio,
            CL_max,
            landing_flare_load_factor,
            rate_of_sink_td,
            approach_alt,
            time_delay,
        ) = inputs.values()
        weight = mass * GRAV_ENGLISH_LBM
        G = GRAV_ENGLISH_GASP

        wing_loading_land = weight / wing_area
        TAS_stall = np.sqrt(2 * wing_loading_land / (CL_max * rho_app))
        TAS_glide = TAS_stall * glide_to_stall_ratio
        theta = np.arcsin(rate_of_sink_max / (60.0 * TAS_glide))
        glide_distance = approach_alt / np.tan(theta)

        # TODO: I didn't understand what is being iterated for glide gamma and it looks like the GASP input/output just uses the maximum rate of sink
        gamma_touchdown = rate_of_sink_td / TAS_glide
        touchdown_velocity_ratio = (glide_to_stall_ratio + 1.0) * 0.5
        RZ = TAS_glide * TAS_glide / G / (landing_flare_load_factor - 1.0)

        flare_alt = (
            TAS_glide
            * TAS_glide
            * (theta * theta - gamma_touchdown * gamma_touchdown)
            / (2.0 * G * (landing_flare_load_factor - 1.0))
        )
        TAS_touchdown = touchdown_velocity_ratio * TAS_stall
        tr_distance = ((RZ * theta) / 2.0) * ((1.0 - gamma_touchdown / theta) ** 2)
        delay_distance = TAS_touchdown * time_delay

        outputs["flare_alt"] = flare_alt
        outputs[Mission.Landing.INITIAL_VELOCITY] = TAS_glide
        outputs[Mission.Landing.STALL_VELOCITY] = TAS_stall
        outputs["TAS_touchdown"] = TAS_touchdown
        outputs["density_ratio"] = rho_app / 0.002378
        outputs["wing_loading_land"] = wing_loading_land
        outputs["glide_distance"] = glide_distance
        outputs["tr_distance"] = tr_distance
        outputs["delay_distance"] = delay_distance
        outputs["theta"] = theta

    def compute_partials(self, inputs, J):
        (
            rho_app,
            rate_of_sink_max,
            mass,
            wing_area,
            glide_to_stall_ratio,
            CL_max,
            landing_flare_load_factor,
            rate_of_sink_td,
            approach_alt,
            time_delay,
        ) = inputs.values()
        weight = mass * GRAV_ENGLISH_LBM
        G = GRAV_ENGLISH_GASP

        wing_loading_land = weight / wing_area
        TAS_stall = np.sqrt(2 * wing_loading_land / (CL_max * rho_app))
        TAS_glide = TAS_stall * glide_to_stall_ratio
        gamma_touchdown = rate_of_sink_td / TAS_glide
        touchdown_velocity_ratio = (glide_to_stall_ratio + 1.0) * 0.5
        theta = np.arcsin(rate_of_sink_max / (60.0 * TAS_glide))
        TAS_touchdown = touchdown_velocity_ratio * TAS_stall

        dTasStall_dWeight = (
            (2 * (1 / wing_area) / (CL_max * rho_app)) ** 0.5 * 0.5 * weight ** (-0.5)
        )
        dTasStall_dWingArea = (
            (2 * weight / (CL_max * rho_app)) ** 0.5 * (-0.5) * wing_area ** (-1.5)
        )
        dTasStall_dClMax = (
            (2 * wing_loading_land / (rho_app)) ** 0.5 * (-0.5) * CL_max ** (-1.5)
        )
        dTasStall_dRhoApp = (
            (2 * wing_loading_land / (CL_max)) ** 0.5 * (-0.5) * rho_app ** (-1.5)
        )
        dTasGlide_dWeight = (
            dTasStall_dWeight * glide_to_stall_ratio
        )
        dTasTd_dWeight = (
            touchdown_velocity_ratio * dTasStall_dWeight
        )
        dTheta_dWeight = (
            (1 - (rate_of_sink_max / (60.0 * TAS_glide)) ** 2) ** (-0.5)
            * (-rate_of_sink_max / (60.0 * TAS_glide**2))
            * dTasGlide_dWeight
        )

        J[Mission.Landing.INITIAL_VELOCITY, Dynamic.Mission.MASS] = \
            dTasGlide_dWeight * GRAV_ENGLISH_LBM
        J[Mission.Landing.INITIAL_VELOCITY, Aircraft.Wing.AREA] = dTasGlide_dWingArea = (
            dTasStall_dWingArea * glide_to_stall_ratio
        )
        J[Mission.Landing.INITIAL_VELOCITY, "CL_max"] = dTasGlide_dClMax = (
            dTasStall_dClMax * glide_to_stall_ratio
        )
        J[Mission.Landing.INITIAL_VELOCITY, "rho_app"] = dTasGlide_dRhoApp = (
            dTasStall_dRhoApp * glide_to_stall_ratio
        )
        J[Mission.Landing.INITIAL_VELOCITY,
            Mission.Landing.GLIDE_TO_STALL_RATIO] = TAS_stall

        J[Mission.Landing.STALL_VELOCITY, Dynamic.Mission.MASS] = \
            dTasStall_dWeight * GRAV_ENGLISH_LBM
        J[Mission.Landing.STALL_VELOCITY, Aircraft.Wing.AREA] = dTasStall_dWingArea
        J[Mission.Landing.STALL_VELOCITY, "CL_max"] = dTasStall_dClMax
        J[Mission.Landing.STALL_VELOCITY, "rho_app"] = dTasStall_dRhoApp

        J["TAS_touchdown", Mission.Landing.GLIDE_TO_STALL_RATIO] = dTasTd_dGlideToStallRatio = (
            0.5 * TAS_stall
        )
        J["TAS_touchdown", Dynamic.Mission.MASS] = dTasTd_dWeight * GRAV_ENGLISH_LBM
        J["TAS_touchdown", Aircraft.Wing.AREA] = dTasTd_dWingArea = (
            touchdown_velocity_ratio * dTasStall_dWingArea
        )
        J["TAS_touchdown", "CL_max"] = dTasTd_dClMax = (
            touchdown_velocity_ratio * dTasStall_dClMax
        )
        J["TAS_touchdown", "rho_app"] = dTasTd_dRhoApp = (
            touchdown_velocity_ratio * dTasStall_dRhoApp
        )

        J["density_ratio", "rho_app"] = 1 / 0.002378

        J["wing_loading_land", Dynamic.Mission.MASS] = GRAV_ENGLISH_LBM / wing_area
        J["wing_loading_land", Aircraft.Wing.AREA] = -weight / wing_area**2

        np.arcsin(rate_of_sink_max / (60.0 * TAS_glide))

        J["theta", Mission.Landing.MAXIMUM_SINK_RATE] = dTheta_dRateOfSinkMax = (
            (1 - (rate_of_sink_max / (60.0 * TAS_glide)) ** 2) ** (-0.5)
            * 1
            / (60.0 * TAS_glide)
        )
        J["theta", Dynamic.Mission.MASS] = dTheta_dWeight * GRAV_ENGLISH_LBM
        J["theta", Aircraft.Wing.AREA] = dTheta_dWingArea = (
            (1 - (rate_of_sink_max / (60.0 * TAS_glide)) ** 2) ** (-0.5)
            * (-rate_of_sink_max / (60.0 * TAS_glide**2))
            * dTasGlide_dWingArea
        )
        J["theta", "CL_max"] = dTheta_dClMax = (
            (1 - (rate_of_sink_max / (60.0 * TAS_glide)) ** 2) ** (-0.5)
            * (-rate_of_sink_max / (60.0 * TAS_glide**2))
            * dTasGlide_dClMax
        )
        J["theta", "rho_app"] = dTheta_dRhoApp = (
            (1 - (rate_of_sink_max / (60.0 * TAS_glide)) ** 2) ** (-0.5)
            * (-rate_of_sink_max / (60.0 * TAS_glide**2))
            * dTasGlide_dRhoApp
        )
        J["theta", Mission.Landing.GLIDE_TO_STALL_RATIO] = dTheta_dGlideToStallRatio = (
            (1 - (rate_of_sink_max / (60.0 * TAS_glide)) ** 2) ** (-0.5)
            * (-rate_of_sink_max / (60.0 * TAS_glide**2))
            * TAS_stall
        )

        approach_alt / np.tan(theta)

        J["glide_distance", Mission.Landing.INITIAL_ALTITUDE] = 1 / np.tan(theta)
        J["glide_distance", Mission.Landing.MAXIMUM_SINK_RATE] = (
            -approach_alt
            / (np.tan(theta)) ** 2
            * (1 / np.cos(theta)) ** 2
            * dTheta_dRateOfSinkMax
        )
        J["glide_distance", Dynamic.Mission.MASS] = (
            -approach_alt
            / (np.tan(theta)) ** 2
            * (1 / np.cos(theta)) ** 2
            * dTheta_dWeight * GRAV_ENGLISH_LBM
        )
        J["glide_distance", Aircraft.Wing.AREA] = (
            -approach_alt
            / (np.tan(theta)) ** 2
            * (1 / np.cos(theta)) ** 2
            * dTheta_dWingArea
        )
        J["glide_distance", "CL_max"] = (
            -approach_alt
            / (np.tan(theta)) ** 2
            * (1 / np.cos(theta)) ** 2
            * dTheta_dClMax
        )
        J["glide_distance", "rho_app"] = (
            -approach_alt
            / (np.tan(theta)) ** 2
            * (1 / np.cos(theta)) ** 2
            * dTheta_dRhoApp
        )
        J["glide_distance", Mission.Landing.GLIDE_TO_STALL_RATIO] = (
            -approach_alt
            / (np.tan(theta)) ** 2
            * (1 / np.cos(theta)) ** 2
            * dTheta_dGlideToStallRatio
        )

        RZ = TAS_glide**2 / G / (landing_flare_load_factor - 1.0)

        dRZ_dWeight = (
            2 * TAS_glide / G / (landing_flare_load_factor - 1.0) * dTasGlide_dWeight
        )
        dRZ_dWingArea = (
            2 * TAS_glide / G / (landing_flare_load_factor - 1.0) * dTasGlide_dWingArea
        )
        dRZ_dClMax = (
            2 * TAS_glide / G / (landing_flare_load_factor - 1.0) * dTasGlide_dClMax
        )
        dRZ_dRhoApp = (
            2 * TAS_glide / G / (landing_flare_load_factor - 1.0) * dTasGlide_dRhoApp
        )
        dRZ_dGlideToStallRatio = (
            2 * TAS_glide / G / (landing_flare_load_factor - 1.0) * TAS_stall
        )
        dRZ_dLandingFlareLoadFactor = (
            -(TAS_glide**2) / G / (landing_flare_load_factor - 1.0) ** 2
        )

        gamma_touchdown = rate_of_sink_td / TAS_glide

        dGammaTd_dRateOfSinkTd = 1 / TAS_glide
        dGammaTd_dWeight = -rate_of_sink_td / TAS_glide**2 * dTasGlide_dWeight
        dGammaTd_dWingArea = -rate_of_sink_td / TAS_glide**2 * dTasGlide_dWingArea
        dGammaTd_dClMax = -rate_of_sink_td / TAS_glide**2 * dTasGlide_dClMax
        dGammaTd_dRhoApp = -rate_of_sink_td / TAS_glide**2 * dTasGlide_dRhoApp
        dGammaTd_dGlideToStallRatio = -rate_of_sink_td / TAS_glide**2 * TAS_stall

        inter1 = (RZ * theta) / 2.0
        dInter1_dLandingFlareLoadFactor = 0.5 * (dRZ_dLandingFlareLoadFactor * theta)
        dInter1_dRateOfSinkMax = 0.5 * (RZ * dTheta_dRateOfSinkMax)
        dInter1_dWeight = 0.5 * (dRZ_dWeight * theta + RZ * dTheta_dWeight)
        dInter1_dWingArea = 0.5 * (dRZ_dWingArea * theta + RZ * dTheta_dWingArea)
        dInter1_dClMax = 0.5 * (dRZ_dClMax * theta + RZ * dTheta_dClMax)
        dInter1_dRhoApp = 0.5 * (dRZ_dRhoApp * theta + RZ * dTheta_dRhoApp)
        dInter1_dGlideToStallRatio = 0.5 * (
            dRZ_dGlideToStallRatio * theta + RZ * dTheta_dGlideToStallRatio
        )

        inter2 = (1.0 - gamma_touchdown / theta) ** 2
        dInter2_dRateOfSinkMax = (
            2
            * (1.0 - gamma_touchdown / theta)
            * gamma_touchdown
            / theta**2
            * dTheta_dRateOfSinkMax
        )
        dInter2_dWeight = (
            -2
            * (1.0 - gamma_touchdown / theta)
            * (theta * dGammaTd_dWeight - gamma_touchdown * dTheta_dWeight)
            / theta**2
        )
        dInter2_dWingArea = (
            -2
            * (1.0 - gamma_touchdown / theta)
            * (theta * dGammaTd_dWingArea - gamma_touchdown * dTheta_dWingArea)
            / theta**2
        )
        dInter2_dClMax = (
            -2
            * (1.0 - gamma_touchdown / theta)
            * (theta * dGammaTd_dClMax - gamma_touchdown * dTheta_dClMax)
            / theta**2
        )
        dInter2_dRhoApp = (
            -2
            * (1.0 - gamma_touchdown / theta)
            * (theta * dGammaTd_dRhoApp - gamma_touchdown * dTheta_dRhoApp)
            / theta**2
        )
        dInter2_dGlideToStallRatio = (
            -2
            * (1.0 - gamma_touchdown / theta)
            * (
                theta * dGammaTd_dGlideToStallRatio
                - gamma_touchdown * dTheta_dGlideToStallRatio
            )
            / theta**2
        )
        dInter2_dRateOfSinkTd = (
            -2 * (1.0 - gamma_touchdown / theta) * dGammaTd_dRateOfSinkTd / theta
        )

        tr_distance = inter1 * inter2

        J["tr_distance", Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR] = (
            dInter1_dLandingFlareLoadFactor * inter2
        )
        J["tr_distance", Mission.Landing.MAXIMUM_SINK_RATE] = (
            dInter1_dRateOfSinkMax * inter2 + inter1 * dInter2_dRateOfSinkMax
        )
        J["tr_distance", Dynamic.Mission.MASS] = (
            dInter1_dWeight * inter2 + inter1 * dInter2_dWeight
        ) * GRAV_ENGLISH_LBM
        J["tr_distance", Aircraft.Wing.AREA] = (
            dInter1_dWingArea * inter2 + inter1 * dInter2_dWingArea
        )
        J["tr_distance", "CL_max"] = dInter1_dClMax * inter2 + inter1 * dInter2_dClMax
        J["tr_distance", "rho_app"] = (
            dInter1_dRhoApp * inter2 + inter1 * dInter2_dRhoApp
        )
        J["tr_distance", Mission.Landing.GLIDE_TO_STALL_RATIO] = (
            dInter1_dGlideToStallRatio * inter2 + inter1 * dInter2_dGlideToStallRatio
        )
        J["tr_distance", Mission.Landing.TOUCHDOWN_SINK_RATE] = inter1 * dInter2_dRateOfSinkTd

        J["delay_distance", Mission.Landing.GLIDE_TO_STALL_RATIO] = (
            time_delay * dTasTd_dGlideToStallRatio
        )
        J["delay_distance", Dynamic.Mission.MASS] = \
            time_delay * dTasTd_dWeight * GRAV_ENGLISH_LBM
        J["delay_distance", Aircraft.Wing.AREA] = time_delay * dTasTd_dWingArea
        J["delay_distance", "CL_max"] = time_delay * dTasTd_dClMax
        J["delay_distance", "rho_app"] = time_delay * dTasTd_dRhoApp
        J["delay_distance", Mission.Landing.BRAKING_DELAY] = TAS_touchdown

        flare_alt = (
            TAS_glide**2
            * (theta**2 - gamma_touchdown**2)
            / (2.0 * G * (landing_flare_load_factor - 1.0))
        )

        J["flare_alt", Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR] = (
            -(TAS_glide**2)
            * (theta**2 - gamma_touchdown**2)
            / (2.0 * G * (landing_flare_load_factor - 1.0)) ** 2
            * 2
            * G
        )
        J["flare_alt", Mission.Landing.TOUCHDOWN_SINK_RATE] = (
            -2
            * TAS_glide**2
            * gamma_touchdown
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * dGammaTd_dRateOfSinkTd
        )
        J["flare_alt", Mission.Landing.MAXIMUM_SINK_RATE] = (
            2
            * TAS_glide**2
            * theta
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * dTheta_dRateOfSinkMax
        )
        J["flare_alt", Dynamic.Mission.MASS] = (
            1
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * (
                2 * TAS_glide * dTasGlide_dWeight * (theta**2 - gamma_touchdown**2)
                + TAS_glide**2
                * (2 * theta * dTheta_dWeight - 2 * gamma_touchdown * dGammaTd_dWeight)
            ) * GRAV_ENGLISH_LBM
        )
        J["flare_alt", Aircraft.Wing.AREA] = (
            1
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * (
                2
                * TAS_glide
                * dTasGlide_dWingArea
                * (theta**2 - gamma_touchdown**2)
                + TAS_glide**2
                * (
                    2 * theta * dTheta_dWingArea
                    - 2 * gamma_touchdown * dGammaTd_dWingArea
                )
            )
        )
        J["flare_alt", "CL_max"] = (
            1
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * (
                2 * TAS_glide * dTasGlide_dClMax * (theta**2 - gamma_touchdown**2)
                + TAS_glide**2
                * (2 * theta * dTheta_dClMax - 2 * gamma_touchdown * dGammaTd_dClMax)
            )
        )
        J["flare_alt", "rho_app"] = (
            1
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * (
                2 * TAS_glide * dTasGlide_dRhoApp * (theta**2 - gamma_touchdown**2)
                + TAS_glide**2
                * (2 * theta * dTheta_dRhoApp - 2 * gamma_touchdown * dGammaTd_dRhoApp)
            )
        )
        J["flare_alt", Mission.Landing.GLIDE_TO_STALL_RATIO] = (
            1
            / (2.0 * G * (landing_flare_load_factor - 1.0))
            * (
                2 * TAS_glide * TAS_stall * (theta**2 - gamma_touchdown**2)
                + TAS_glide**2
                * (
                    2 * theta * dTheta_dGlideToStallRatio
                    - 2 * gamma_touchdown * dGammaTd_dGlideToStallRatio
                )
            )
        )


class LandingGroundRollComponent(om.ExplicitComponent):
    def setup(self):

        self.add_input("touchdown_CD", val=0.0, units='unitless',
                       desc="CDRL: CD at touchdown")
        self.add_input("touchdown_CL", val=0.0, units='unitless',
                       desc="CLRL: CL at touchdown")

        add_aviary_input(self, Mission.Landing.STALL_VELOCITY, val=0.0)
        self.add_input(
            "TAS_touchdown", val=0.0, units="ft/s", desc="VTD: velocity at touchdown"
        )
        self.add_input(
            "thrust_idle",
            val=0.0,
            units="lbf",
            desc="TIDLE: idle thrust at start of landing",
        )
        self.add_input(
            "density_ratio", val=0.0, units="unitless", desc="DRAT: density ratio for DLAND"
        )
        self.add_input(
            "wing_loading_land",
            val=0.0,
            units="lbf/ft**2",
            desc="WOS: wing loading at landing",
        )
        self.add_input(
            "glide_distance", val=0.0, units="ft", desc="DLGL: glide distance"
        )
        self.add_input(
            "tr_distance",
            val=0.0,
            units="ft",
            desc="DLTR: distance from glide to touchdown",
        )
        self.add_input(
            "delay_distance",
            val=0.0,
            units="ft",
            desc="DDELAY: touchdown to brake application",
        )
        self.add_input(
            "CL_max", val=0.0, units="unitless", desc="CLMX: max CL at approach altitude"
        )
        self.add_input(
            Dynamic.Mission.MASS,
            val=0.0,
            units="lbm",
            desc="WL: aircraft mass at start of landing",
        )

        self.add_output(
            "ground_roll_distance",
            val=0.0,
            units="ft",
            desc="DLG: distance during braked ground roll",
        )
        add_aviary_output(self, Mission.Landing.GROUND_DISTANCE, val=0.0)
        self.add_output("average_acceleration", val=0.0, units="unitless",  # renamed from GASP var ABAR
                        desc="average acceleration/decelleration based on starting speed (TAS) and rollout distance")

        self.declare_partials(
            "ground_roll_distance",
            [
                "wing_loading_land",
                "density_ratio",
                "touchdown_CD",
                "touchdown_CL",
                "thrust_idle",
                Dynamic.Mission.MASS,
                "CL_max",
                Mission.Landing.STALL_VELOCITY,
                "TAS_touchdown",
            ],
        )
        self.declare_partials(
            Mission.Landing.GROUND_DISTANCE,
            [
                "wing_loading_land",
                "density_ratio",
                "touchdown_CD",
                "touchdown_CL",
                "thrust_idle",
                Dynamic.Mission.MASS,
                "CL_max",
                Mission.Landing.STALL_VELOCITY,
                "TAS_touchdown",
                "tr_distance",
                "delay_distance",
                "glide_distance",
            ],
        )
        self.declare_partials(
            "average_acceleration",
            [
                "TAS_touchdown",
                "wing_loading_land",
                "density_ratio",
                "touchdown_CD",
                "touchdown_CL",
                "thrust_idle",
                Dynamic.Mission.MASS,
                "CL_max",
                Mission.Landing.STALL_VELOCITY,
            ],
        )

    def compute(self, inputs, outputs):
        (
            touchdown_CD,
            touchdown_CL,
            TAS_stall,
            TAS_touchdown,
            thrust_idle,
            density_ratio,
            wing_loading_land,
            glide_distance,
            tr_distance,
            delay_distance,
            CL_max,
            mass,
        ) = inputs.values()

        weight = mass * GRAV_ENGLISH_LBM
        G = GRAV_ENGLISH_GASP
        MUB = MU_LANDING

        DLRL = touchdown_CD - (MUB * touchdown_CL)
        ARAT = DLRL / (CL_max * (TAS_stall / TAS_touchdown) ** 2)
        thrust_to_landing_weight_ratio = thrust_idle / weight
        FRAT = MUB - thrust_to_landing_weight_ratio
        ALN = np.log(FRAT / (FRAT + ARAT))

        ground_roll_distance = (
            -13.0287 * wing_loading_land * ALN / (density_ratio * DLRL)
        )
        total_distance = (
            ground_roll_distance + tr_distance + delay_distance + glide_distance
        )
        average_acceleration = TAS_touchdown**2.0 / (ground_roll_distance * 2.0 * G)

        outputs["ground_roll_distance"] = ground_roll_distance
        outputs[Mission.Landing.GROUND_DISTANCE] = total_distance
        outputs["average_acceleration"] = average_acceleration

    def compute_partials(self, inputs, J):
        (
            touchdown_CD,
            touchdown_CL,
            TAS_stall,
            TAS_touchdown,
            thrust_idle,
            density_ratio,
            wing_loading_land,
            glide_distance,
            tr_distance,
            delay_distance,
            CL_max,
            mass,
        ) = inputs.values()
        weight = mass * GRAV_ENGLISH_LBM
        G = GRAV_ENGLISH_GASP
        MUB = MU_LANDING

        DLRL = touchdown_CD - (MUB * touchdown_CL)
        ARAT = DLRL / (CL_max * (TAS_stall / TAS_touchdown) ** 2)
        thrust_to_landing_weight_ratio = thrust_idle / weight
        FRAT = MUB - thrust_to_landing_weight_ratio
        ALN = np.log(FRAT / (FRAT + ARAT))

        dDLRL_dTouchdownCD = 1
        dDLRL_dTouchdownCL = -MUB

        dARAT_dTouchdownCD = dDLRL_dTouchdownCD / (
            CL_max * (TAS_stall / TAS_touchdown) ** 2
        )
        dARAT_dTouchdownCL = dDLRL_dTouchdownCL / (
            CL_max * (TAS_stall / TAS_touchdown) ** 2
        )
        dARAT_dClMax = -DLRL / (CL_max**2 * (TAS_stall / TAS_touchdown) ** 2)
        dARAT_dTasStall = -2 * DLRL / (CL_max * (TAS_stall**3 / TAS_touchdown**2))
        dARAT_dTasTouchdown = 2 * TAS_touchdown * DLRL / (CL_max * (TAS_stall) ** 2)

        dFRAT_dThrustIdle = -1 / weight
        dFRAT_dWeight = thrust_idle / weight**2

        dALN_dTouchdownCD = (
            1
            / (FRAT / (FRAT + ARAT))
            * (-FRAT / (FRAT + ARAT) ** 2)
            * dARAT_dTouchdownCD
        )
        dALN_dTouchdownCL = (
            1
            / (FRAT / (FRAT + ARAT))
            * (-FRAT / (FRAT + ARAT) ** 2)
            * dARAT_dTouchdownCL
        )
        dALN_dThrustIdle = (
            1
            / (FRAT / (FRAT + ARAT))
            * ((FRAT + ARAT) - FRAT)
            / (FRAT + ARAT) ** 2
            * dFRAT_dThrustIdle
        )
        dALN_dWeight = (
            1
            / (FRAT / (FRAT + ARAT))
            * ((FRAT + ARAT) - FRAT)
            / (FRAT + ARAT) ** 2
            * dFRAT_dWeight
        )
        dALN_dClMax = (
            1 / (FRAT / (FRAT + ARAT)) * (-FRAT / (FRAT + ARAT) ** 2 * dARAT_dClMax)
        )
        dALN_dTasStall = (
            1 / (FRAT / (FRAT + ARAT)) * (-FRAT / (FRAT + ARAT) ** 2 * dARAT_dTasStall)
        )
        dALN_dTasTouchdown = (
            1
            / (FRAT / (FRAT + ARAT))
            * (-FRAT / (FRAT + ARAT) ** 2 * dARAT_dTasTouchdown)
        )

        ground_roll_distance = (
            -13.0287 * wing_loading_land * ALN / (density_ratio * DLRL)
        )
        dGRD_dWeight = (
            -13.0287 * wing_loading_land * dALN_dWeight / (density_ratio * DLRL)
        )

        J["ground_roll_distance", "wing_loading_land"] = dGRD_dWingLoadingLand = (
            -13.0287 * ALN / (density_ratio * DLRL)
        )
        J["ground_roll_distance", "density_ratio"] = dGRD_dDensityRatio = (
            13.0287 * wing_loading_land * ALN / (density_ratio**2 * DLRL)
        )
        J["ground_roll_distance", "touchdown_CD"] = dGRD_dTouchdownCD = (
            -13.0287
            * (wing_loading_land / density_ratio)
            * (DLRL * dALN_dTouchdownCD - ALN * dDLRL_dTouchdownCD)
            / DLRL**2
        )
        J["ground_roll_distance", "touchdown_CL"] = dGRD_dTouchdownCL = (
            -13.0287
            * (wing_loading_land / density_ratio)
            * (DLRL * dALN_dTouchdownCL - ALN * dDLRL_dTouchdownCL)
            / DLRL**2
        )
        J["ground_roll_distance", "thrust_idle"] = dGRD_dThrustIdle = (
            -13.0287 * wing_loading_land * dALN_dThrustIdle / (density_ratio * DLRL)
        )
        J["ground_roll_distance", Dynamic.Mission.MASS] = dGRD_dWeight * GRAV_ENGLISH_LBM
        J["ground_roll_distance", "CL_max"] = dGRD_dClMax = (
            -13.0287 * wing_loading_land * dALN_dClMax / (density_ratio * DLRL)
        )
        J["ground_roll_distance", Mission.Landing.STALL_VELOCITY] = dGRD_dTasStall = (
            -13.0287 * wing_loading_land * dALN_dTasStall / (density_ratio * DLRL)
        )
        J["ground_roll_distance", "TAS_touchdown"] = dGRD_dTasTouchdown = (
            -13.0287 * wing_loading_land * dALN_dTasTouchdown / (density_ratio * DLRL)
        )

        J[Mission.Landing.GROUND_DISTANCE,
            "wing_loading_land"] = dGRD_dWingLoadingLand
        J[Mission.Landing.GROUND_DISTANCE, "density_ratio"] = dGRD_dDensityRatio
        J[Mission.Landing.GROUND_DISTANCE, "touchdown_CD"] = dGRD_dTouchdownCD
        J[Mission.Landing.GROUND_DISTANCE, "touchdown_CL"] = dGRD_dTouchdownCL
        J[Mission.Landing.GROUND_DISTANCE, "thrust_idle"] = dGRD_dThrustIdle
        J[Mission.Landing.GROUND_DISTANCE, Dynamic.Mission.MASS] = \
            dGRD_dWeight * GRAV_ENGLISH_LBM
        J[Mission.Landing.GROUND_DISTANCE, "CL_max"] = dGRD_dClMax
        J[Mission.Landing.GROUND_DISTANCE,
            Mission.Landing.STALL_VELOCITY] = dGRD_dTasStall
        J[Mission.Landing.GROUND_DISTANCE, "TAS_touchdown"] = dGRD_dTasTouchdown
        J[Mission.Landing.GROUND_DISTANCE, "tr_distance"] = 1
        J[Mission.Landing.GROUND_DISTANCE, "delay_distance"] = 1
        J[Mission.Landing.GROUND_DISTANCE, "glide_distance"] = 1

        J["average_acceleration", "wing_loading_land"] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dWingLoadingLand
        )
        J["average_acceleration", "density_ratio"] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dDensityRatio
        )
        J["average_acceleration", "touchdown_CD"] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dTouchdownCD
        )
        J["average_acceleration", "touchdown_CL"] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dTouchdownCL
        )
        J["average_acceleration", "thrust_idle"] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dThrustIdle
        )
        J["average_acceleration", Dynamic.Mission.MASS] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dWeight * GRAV_ENGLISH_LBM
        )
        J["average_acceleration", "CL_max"] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dClMax
        )
        J["average_acceleration", Mission.Landing.STALL_VELOCITY] = (
            -(TAS_touchdown**2.0)
            / (ground_roll_distance**2 * 2.0 * G)
            * dGRD_dTasStall
        )
        J["average_acceleration", "TAS_touchdown"] = (
            ground_roll_distance * 2 * G * 2 * TAS_touchdown
            - TAS_touchdown**2 * 2 * G * dGRD_dTasTouchdown
        ) / (ground_roll_distance * 2.0 * G) ** 2



import unittest

from openmdao.utils.assert_utils import assert_near_equal

from aviary.subsystems.mass.flops_based.distributed_prop import (
    distributed_engine_count_factor, distributed_nacelle_diam_factor,
    distributed_nacelle_diam_factor_deriv, distributed_thrust_factor)


class DistributedPropulsionFactorsTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_case_dist_eng_count(self):
        tol = 1e-12

        for num_eng in [1, 2, 3, 4]:
            num_eng_fact = distributed_engine_count_factor(num_eng)
            assert_near_equal(num_eng_fact, num_eng, tol)

        num_eng = 5
        num_eng_fact = distributed_engine_count_factor(num_eng)
        assert_near_equal(num_eng_fact, 4.643501108793284, tol)

    def test_case_dist_thrust(self):
        tol = 1e-12
        max_thrust = 10000.0

        for num_eng in [1, 2, 3, 4]:
            thrust_fact = distributed_thrust_factor(max_thrust, num_eng)
            assert_near_equal(thrust_fact, max_thrust/num_eng, tol)

        num_eng = 5
        thrust_fact = distributed_thrust_factor(max_thrust, num_eng)
        assert_near_equal(thrust_fact, max_thrust/4.643501108793284, tol)

    def test_case_dist_nacelle_diam(self):
        tol = 1e-12
        diam_nacelle = 10.0

        for num_eng in [1, 2, 3, 4]:
            diam_fact = distributed_nacelle_diam_factor(diam_nacelle, num_eng)
            assert_near_equal(diam_fact, diam_nacelle, tol)

        num_eng = 5
        diam_fact = distributed_nacelle_diam_factor(diam_nacelle, num_eng)
        assert_near_equal(diam_fact, diam_nacelle*1.11803398875, tol)

    def test_case_dist_nacelle_diam_deriv(self):
        tol = 1e-12

        for num_eng in [1, 2, 3, 4]:
            diam_fact = distributed_nacelle_diam_factor_deriv(num_eng)
            assert_near_equal(diam_fact, 1.0, tol)

        num_eng = 5
        diam_fact = distributed_nacelle_diam_factor_deriv(num_eng)
        assert_near_equal(diam_fact, 1.11803398875, tol)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.mission.gasp_based.ode.accel_ode import AccelODE
from aviary.mission.gasp_based.ode.ascent_ode import AscentODE
from aviary.mission.gasp_based.ode.climb_ode import ClimbODE
from aviary.mission.gasp_based.ode.descent_ode import DescentODE
from aviary.mission.gasp_based.ode.flight_path_ode import FlightPathODE
from aviary.mission.gasp_based.phases.landing_group import LandingSegment
from aviary.mission.gasp_based.ode.groundroll_ode import GroundrollODE
from aviary.mission.gasp_based.ode.rotation_ode import RotationODE
from aviary.mission.gasp_based.ode.time_integration_base_classes import SimuPyProblem
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.enums import AlphaModes, AnalysisScheme, SpeedType
from aviary.variable_info.variables import Aircraft, Dynamic, Mission

DEBUG = 0


class SGMGroundroll(SimuPyProblem):
    '''
    This creates a subproblem for the groundroll phase of the trajectory that will
    be solved using SGM.
    Groundroll ends when TAS reaches rotation speed.
    '''

    def __init__(
        self,
        phase_name='groundroll',
        VR_value=143.1,
        VR_units="kn",
        ode_args={},
        simupy_args={},
    ):

        super().__init__(
            GroundrollODE(analysis_scheme=AnalysisScheme.SHOOTING, **ode_args),
            outputs=["normal_force"],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
                Dynamic.Mission.VELOCITY,
            ],
            # state_units=['lbm','nmi','ft','ft/s'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            **simupy_args,
        )

        self.phase_name = phase_name
        self.VR_value = VR_value
        self.VR_units = VR_units
        self.event_channel_names = [Dynamic.Mission.VELOCITY]
        self.num_events = len(self.event_channel_names)

    def event_equation_function(self, t, x):
        self.time = t
        self.state = x
        return self.get_val("velocity", units='ft/s') - self.VR_value


class SGMRotation(SimuPyProblem):
    '''
    This creates a subproblem for the rotation phase of the trajectory that will
    be solved using SGM.
    Rotation ends when the normal force on the runway reaches 0.
    '''

    def __init__(
        self,
        phase_name='rotation',
        ode_args={},
        simupy_args={},
    ):
        super().__init__(
            RotationODE(analysis_scheme=AnalysisScheme.SHOOTING, **ode_args),
            outputs=["normal_force", "alpha"],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
            ],
            # state_units=['lbm','nmi','ft'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            **simupy_args,
        )

        self.phase_name = phase_name
        self.event_channel_names = ["normal_force"]
        self.num_events = len(self.event_channel_names)

    def event_equation_function(self, t, x):
        self.output_equation_function(t, x)
        self.compute()
        norm_force = self.get_val("normal_force", units="lbf") + 0.0

        return norm_force


# TODO : turn these into parameters? inputs? they need to match between
# ODE and SimuPy wrappers
load_factor_max = 1.10
velocity_rate_safety = -np.inf  # 100.
fuselage_pitch_max = 15.0
gear_retraction_alt = 50.0
flap_retraction_alt = 400.0
ascent_termination_alt = 500.0


class SGMAscent(SimuPyProblem):
    '''
    This creates a subproblem for the ascent phase of the trajectory that will
    be solved using SGM.
    Ascent ends at ascent_termination_alt and retracts the gear and flaps at their
    respective retraction altitudes.
    '''

    def __init__(
        self,
        phase_name='ascent',
        alpha_mode=AlphaModes.DEFAULT,
        ode_args={},
        simupy_args={},
    ):
        controls = None
        super().__init__(
            AscentODE(analysis_scheme=AnalysisScheme.SHOOTING,
                      alpha_mode=alpha_mode, **ode_args),
            outputs=[
                "load_factor",
                "fuselage_pitch",
                "normal_force",
                "alpha",
            ],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
            ],
            # state_units=['lbm','nmi','ft'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            controls=controls,
            **simupy_args,
        )

        self.phase_name = phase_name
        self.event_channel_names = [
            Dynamic.Mission.ALTITUDE,
            Dynamic.Mission.ALTITUDE,
            Dynamic.Mission.ALTITUDE,
        ]
        self.num_events = len(self.event_channel_names)

        self.event_names = [
            "termination",
            "flaps",
            "gear",
        ]

    def event_equation_function(self, t, x):
        alpha = self.get_alpha(t, x)
        self.ode0.set_val("alpha", alpha)
        self.ode0.output_equation_function(t, x)
        alt = self.ode0.get_val(Dynamic.Mission.ALTITUDE).squeeze()
        return np.array(
            [
                alt - ascent_termination_alt,
                alt - flap_retraction_alt,
                alt - gear_retraction_alt,
            ]
        ).squeeze()

    def update_equation_function(self, t, x, event_channels=None):
        if 0 in event_channels:
            self.output_nan = True
            return x
        elif 1 in event_channels:
            if DEBUG:
                print("flaps!", t)
            self.set_val("t_init_flaps", t)
        elif 2 in event_channels:
            if DEBUG:
                print("gear!", t)
            self.set_val("t_init_gear", t)
        else:
            return np.nan * np.ones(self.dim_state)
        return x


class SGMAscentCombined(SGMAscent):
    '''
    This combines the different methods of limiting angle of attack to ensure that
    none of the constraints are violated.
    '''

    def __init__(
        self,
        phase_name='ascent_combined',
        fuselage_pitch_max=0,
        ode_args={},
        simupy_args={},
    ):
        self.ode_args = ode_args
        super().__init__(alpha_mode=AlphaModes.DEFAULT, ode_args=ode_args, simupy_args=simupy_args)

        self.phase_name = phase_name
        self.fuselage_pitch_max = fuselage_pitch_max

        ode0 = SGMAscent(alpha_mode=AlphaModes.DEFAULT,
                         ode_args=ode_args, simupy_args=simupy_args)
        rotation = SGMAscent(alpha_mode=AlphaModes.ROTATION,
                             ode_args=ode_args, simupy_args=simupy_args)
        load_factor = SGMAscent(alpha_mode=AlphaModes.LOAD_FACTOR,
                                ode_args=ode_args, simupy_args=simupy_args)
        fuselage_pitch = SGMAscent(
            alpha_mode=AlphaModes.FUSELAGE_PITCH, ode_args=ode_args, simupy_args=simupy_args)
        decel = SGMAscent(alpha_mode=AlphaModes.DECELERATION,
                          ode_args=ode_args, simupy_args=simupy_args)

        self.odes = (ode0, rotation, load_factor, fuselage_pitch, decel,)
        (
            self.ode0,
            self.rotation,
            self.load_factor,
            self.fuselage_pitch,
            self.decel
        ) = (
            ode0, rotation, load_factor, fuselage_pitch, decel,
        )

        self.set_val("t_init_flaps", 500.0, units='s')
        self.set_val("t_init_gear", 500.0, units='s')

        self.ode_name = {
            ode0: "ode0",
            rotation: "rotation",
            load_factor: "load_factor",
            fuselage_pitch: "fuselage pitch",
            decel: "decel"
        }

    def prepare_to_integrate(self, t0, x0):
        self.output_nan = False
        self.alpha_cache = {}
        self.prob_cache = {}
        self.last_prob = self.rotation
        for ode in self.odes[:]:
            ode.prepare_to_integrate(t0, x0)
        return self.output_equation_function(t0, x0)
        # or can this just be done with prepare to integrate??

    def set_val(self, *args, **kwargs):
        for ode in self.odes:
            ode.set_val(*args, **kwargs)

    def compute_alpha(self, ode, t, x):
        return ode.output_equation_function(t, x)[list(ode.outputs.keys()).index("alpha")]

    def get_alpha(self, t, x):
        a_key = (t,) + tuple(x)
        # TODO: I think there's a pythonic way to do this
        if a_key in self.alpha_cache:
            alpha = self.alpha_cache[a_key]
        else:
            # in deriv.f, alpha from previous time-step is known and used as seed
            # assume scheduled increment on alpha
            # then check fuselage pitch, clip to max
            # then check load factor, line search -alpha until satisfied
            # then check end SPEED -- keep tas_rate = 0 if tas > vend
            # (not implemented in gaspy)
            # then check decel, line search -alpha until satisfied
            (ode0, rotation, load_factor, fuselage_pitch, decel,) = self.odes
            ode = self.last_prob
            SATISFIED_CONSTRAINTS = False
            for count in range(4):
                alpha = self.compute_alpha(ode, t, x)
                load_factor_val = ode.get_val("load_factor")
                fuselage_pitch_val = ode.get_val("fuselage_pitch", units="deg")
                velocity_rate_val = ode.get_val("velocity_rate")

                if (
                    (load_factor_val > load_factor_max) and not
                    np.isclose(load_factor_val, load_factor_max)
                ):
                    print('*'*20, 'switching to load_factor', '*'*20)
                    ode = load_factor
                    continue
                elif (
                    (fuselage_pitch_val > fuselage_pitch_max) and not
                    np.isclose(fuselage_pitch_val, fuselage_pitch_max)
                ):
                    print('*'*20, 'switching to fuselage_pitch', '*'*20)
                    ode = fuselage_pitch
                    continue
                elif (
                    (velocity_rate_val < velocity_rate_safety) and not
                    np.isclose(velocity_rate_val, velocity_rate_safety)
                ):
                    print('*'*20, 'switching to decel', '*'*20)
                    ode = decel
                    continue
                else:
                    if (
                        np.isnan(load_factor_val) or
                        np.isnan(fuselage_pitch_val) or
                        np.isnan(velocity_rate_val)
                    ):
                        continue
                    SATISFIED_CONSTRAINTS = True
                    break
            if SATISFIED_CONSTRAINTS:
                self.alpha_cache[a_key] = alpha
                self.prob_cache[a_key] = ode
                self.last_prob = ode
            else:
                print("time :", t)
                print("ode :", self.ode_name[ode])
                for key in ["load_factor", "fuselage_pitch", "velocity_rate"]:
                    print(key, ":", ode.get_val(key))
                raise ValueError("Ascent could not satisfy all constraints")

        return alpha

    def get_prob(self, t, x):
        a_key = (t,) + tuple(x)
        if a_key not in self.prob_cache:
            self.get_alpha(t, x)
        return self.prob_cache[a_key]

    def state_equation_function(self, t, x, u=None):
        if np.any(np.isnan(x)):
            return np.ones(self.dim_output) * np.nan
        alpha = self.get_alpha(t, x)
        prob = self.get_prob(t, x)
        prob.set_val("alpha", alpha)
        return prob.state_equation_function(t, x)

    @property
    def compute_totals(self):
        return self.get_prob(self.time, self.state).compute_totals

    def output_equation_function(self, t, x):
        if np.any(np.isnan(x)) or self.output_nan:
            return np.ones(self.dim_output) * np.nan
        alpha = self.get_alpha(t, x)
        prob = self.get_prob(t, x)
        # I think always need to use ode0 and set alpha --
        # caching source problem doesn't necessarily save anything since you don't know
        # it was the last compute. but maybe it's worth a shot?
        # using solver may introduce slight variations depending on how it's walking or
        # not? and need to have a real compute before compute totals - or does that mean
        # use problem?
        prob.set_val("alpha", alpha)
        self.time = t
        self.state = x
        return prob.output_equation_function(t, x)


class SGMAccel(SimuPyProblem):
    '''
    This creates a subproblem for the acceleration phase of the trajectory that will
    be solved using SGM
    '''

    def __init__(
        self,
        phase_name='accel',
        VC_value=250.0,
        VC_units="kn",
        ode_args={},
        simupy_args={},
    ):
        ode = AccelODE(analysis_scheme=AnalysisScheme.SHOOTING, **ode_args)
        super().__init__(
            ode,
            outputs=["EAS", "mach", "alpha"],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
            ],
            # state_units=['lbm','nmi','ft'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            **simupy_args,
        )
        self.phase_name = phase_name
        self.VC_value = VC_value
        self.VC_units = VC_units
        self.event_channel_names = [
            "EAS",
        ]
        self.num_events = len(self.event_channel_names)

    def event_equation_function(self, t, x):
        self.output_equation_function(t, x)
        return self.get_val("EAS", units=self.VC_units) - self.VC_value


class SGMClimb(SimuPyProblem):
    '''
    This creates a subproblem for the climb phase of the trajectory that will
    be solved using SGM
    '''

    def __init__(
        self,
        phase_name='climb',
        input_speed_type=SpeedType.EAS,
        input_speed_units="kn",
        speed_trigger_units=None,
        alt_trigger_units="ft",
        ode_args={},
        simupy_args={},
    ):
        ode = ClimbODE(
            analysis_scheme=AnalysisScheme.SHOOTING,
            input_speed_type=input_speed_type,
            speed_trigger_units=speed_trigger_units,
            alt_trigger_units=alt_trigger_units,
            **ode_args,
        )
        self.input_speed_type = input_speed_type
        self.input_speed_units = input_speed_units
        if input_speed_type is SpeedType.EAS:
            self.speed_trigger_name = "mach"
        elif input_speed_type is SpeedType.MACH:
            self.speed_trigger_name = "TAS_violation"
        else:
            raise ValueError("bad speed type")
        self.speed_trigger_units = speed_trigger_units
        self.alt_trigger_units = alt_trigger_units
        super().__init__(
            ode,
            outputs=[
                "alpha",
                Dynamic.Mission.FLIGHT_PATH_ANGLE,
                "required_lift",
                "lift",
                "mach",
                "EAS",
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.THRUST_TOTAL,
                "drag",
                Dynamic.Mission.ALTITUDE_RATE,
            ],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
            ],
            # state_units=['lbm','nmi','ft'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            **simupy_args,
        )
        self.phase_name = phase_name
        self.event_channel_names = [
            Dynamic.Mission.ALTITUDE,
            self.speed_trigger_name,
        ]
        self.num_events = len(self.event_channel_names)

    def event_equation_function(self, t, x):
        self.output_equation_function(t, x)
        alt = self.get_val(Dynamic.Mission.ALTITUDE,
                           units=self.alt_trigger_units).squeeze()
        alt_trigger = self.get_val("alt_trigger", units=self.alt_trigger_units).squeeze()

        speed = self.get_val(
            self.speed_trigger_name, units=self.speed_trigger_units
        ).squeeze()
        speed_trigger = self.get_val(
            "speed_trigger", units=self.speed_trigger_units
        ).squeeze()
        return np.array([alt - alt_trigger, speed - speed_trigger])


class SGMCruise(SimuPyProblem):
    '''
    This creates a subproblem for the cruise phase of the trajectory that will
    be solved using SGM
    '''

    def __init__(
        self,
        phase_name='cruise',
        alpha_mode=AlphaModes.DEFAULT,
        input_speed_type=SpeedType.MACH,
        input_speed_units="kn",
        distance_trigger_units='ft',
        ode_args={},
        simupy_args={},
    ):
        ode = FlightPathODE(
            analysis_scheme=AnalysisScheme.SHOOTING,
            alpha_mode=alpha_mode,
            input_speed_type=input_speed_type,
            clean=True,
            **ode_args,)

        self.distance_trigger_units = distance_trigger_units

        super().__init__(
            ode,
            outputs=[
                "alpha",  # ?
                "lift",
                "EAS",
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.THRUST_TOTAL,
                "drag",
                Dynamic.Mission.ALTITUDE_RATE,
            ],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
            ],
            # state_units=['lbm','nmi','ft'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            **simupy_args,
        )

        self.phase_name = phase_name
        self.event_channel_names = [
            Dynamic.Mission.MASS,
            Dynamic.Mission.DISTANCE,
        ]
        self.num_events = len(self.event_channel_names)

    def event_equation_function(self, t, x):
        self.output_equation_function(t, x)
        current_mass = self.get_val(Dynamic.Mission.MASS, units="lbm").squeeze()
        mass_trigger = self.get_val('mass_trigger.mass_trigger', units="lbm").squeeze()

        distance = self.get_val(Dynamic.Mission.DISTANCE,
                                units=self.distance_trigger_units).squeeze()
        distance_trigger = self.get_val(
            "distance_trigger", units=self.distance_trigger_units).squeeze()

        return np.array([
            current_mass - mass_trigger,
            distance - distance_trigger
        ])


class SGMDescent(SimuPyProblem):
    '''
    This creates a subproblem for the descent phase of the trajectory that will
    be solved using SGM
    '''

    def __init__(
        self,
        phase_name='descent',
        input_speed_type=SpeedType.EAS,
        input_speed_units="kn",
        speed_trigger_units=None,
        alt_trigger_units="ft",
        ode_args={},
        simupy_args={},
    ):
        ode = DescentODE(
            analysis_scheme=AnalysisScheme.SHOOTING,
            input_speed_type=input_speed_type,
            speed_trigger_units=speed_trigger_units,
            alt_trigger_units=alt_trigger_units,
            **ode_args,
        )
        self.input_speed_type = input_speed_type
        self.input_speed_units = input_speed_units
        if input_speed_type is SpeedType.MACH:
            self.speed_trigger_name = "EAS"
        elif input_speed_type is SpeedType.EAS:
            self.speed_trigger_name = "TAS_violation"
        else:
            raise ValueError("bad speed type")
        self.speed_trigger_units = speed_trigger_units
        self.alt_trigger_units = alt_trigger_units
        super().__init__(
            ode,
            outputs=[
                "alpha",
                "required_lift",
                "lift",
                "EAS",
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.THRUST_TOTAL,
                "drag",
                Dynamic.Mission.ALTITUDE_RATE,
            ],
            states=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.DISTANCE,
                Dynamic.Mission.ALTITUDE,
            ],
            # state_units=['lbm','nmi','ft'],
            alternate_state_rate_names={
                Dynamic.Mission.MASS: Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL},
            **simupy_args,
        )

        self.phase_name = phase_name
        self.event_channel_names = [
            Dynamic.Mission.ALTITUDE,
            self.speed_trigger_name,
        ]
        self.num_events = len(self.event_channel_names)

    def event_equation_function(self, t, x):
        self.output_equation_function(t, x)
        alt = self.get_val(Dynamic.Mission.ALTITUDE,
                           units=self.alt_trigger_units).squeeze()
        alt_trigger = self.get_val("alt_trigger", units=self.alt_trigger_units).squeeze()

        speed = self.get_val(
            self.speed_trigger_name, units=self.speed_trigger_units
        ).squeeze()
        speed_trigger = self.get_val(
            "speed_trigger", units=self.speed_trigger_units
        ).squeeze()
        return np.array([alt - alt_trigger, speed - speed_trigger])



import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.constants import GRAV_ENGLISH_LBM, RHO_SEA_LEVEL_METRIC
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class StallSpeed(om.ExplicitComponent):
    def setup(self):
        """
        Calculates the stall speed of the aircraft.
        """

        self.add_input(
            "mass",
            val=150_000,
            units="lbm",
            desc="mass of the aircraft",
        )

        self.add_input("rho", val=1.225, units="kg/m**3", desc="atmospheric density")

        self.add_input("planform_area", val=7, units="m**2", desc="area of the wings")

        self.add_input(
            "Cl_max",
            val=2,
            units='unitless',
            desc="maximum lift coefficient",
        )

        self.add_output(
            "v_stall",
            val=0.1,
            units="m/s",
            desc="stall velocity",
        )

        self.declare_partials("v_stall", "*")

    def compute(self, inputs, outputs):

        weight = inputs["mass"] * GRAV_ENGLISH_LBM
        # # convert from pounds to newtons.
        # This is only necessary because the equation expects newtons,
        # but the mission expects pounds mass instead of pounds force.
        weight = weight*4.44822
        rho = inputs["rho"]
        S = inputs["planform_area"]
        Cl_max = inputs["Cl_max"]

        v_stall = (2 * weight / (rho * S * Cl_max)) ** 0.5

        outputs["v_stall"] = v_stall

    def compute_partials(self, inputs, J):

        weight = inputs["mass"] * GRAV_ENGLISH_LBM
        rho = inputs["rho"]
        S = inputs["planform_area"]
        Cl_max = inputs["Cl_max"]

        rad = 2 * weight / (rho * S * Cl_max)

        J["v_stall", "mass"] = 0.5 * 4.44822**.5 * \
            rad ** (-0.5) * 2 * GRAV_ENGLISH_LBM / (rho * S * Cl_max)
        J["v_stall", "rho"] = (
            0.5 * 4.44822**.5 * rad ** (-0.5) * (-2 * weight) / (rho ** 2 * S * Cl_max)
        )
        J["v_stall", "planform_area"] = (
            0.5 * 4.44822**.5 * rad ** (-0.5) * (-2 * weight) / (rho * S ** 2 * Cl_max)
        )
        J["v_stall", "Cl_max"] = (
            0.5 * 4.44822**.5 * rad ** (-0.5) * (-2 * weight) / (rho * S * Cl_max ** 2)
        )


class FinalTakeoffConditions(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_engines", desc="number of engines on aircraft")

    def setup(self):

        self.add_input(
            "v_stall",
            val=0.1,
            units="m/s",
            desc="stall speed of the aircraft",
        )

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=150_000)

        add_aviary_input(self, Mission.Takeoff.FUEL_SIMPLE, val=10.e3)

        self.add_input("rho", val=1.225, units="kg/m**3", desc="atmospheric density")

        add_aviary_input(self, Aircraft.Wing.AREA, val=7)

        add_aviary_input(self, Mission.Takeoff.LIFT_COEFFICIENT_MAX, val=2)

        add_aviary_input(self, Mission.Design.THRUST_TAKEOFF_PER_ENG, val=100_000)

        add_aviary_input(self, Mission.Takeoff.LIFT_OVER_DRAG, val=2)

        add_aviary_output(self, Mission.Takeoff.GROUND_DISTANCE, val=0)

        add_aviary_output(self, Mission.Takeoff.FINAL_VELOCITY,
                          val=0, units="m/s")

        add_aviary_output(self, Mission.Takeoff.FINAL_MASS, val=0)

        add_aviary_output(self, Mission.Takeoff.FINAL_ALTITUDE, val=0)

    def setup_partials(self):

        self.declare_partials(
            Mission.Takeoff.FINAL_VELOCITY,
            "v_stall",
            val=1.2309,
        )
        self.declare_partials(
            Mission.Takeoff.GROUND_DISTANCE,
            [
                Mission.Design.GROSS_MASS,
                "rho",
                Aircraft.Wing.AREA,
                Mission.Takeoff.LIFT_COEFFICIENT_MAX,
                Mission.Design.THRUST_TAKEOFF_PER_ENG,
                Mission.Takeoff.LIFT_OVER_DRAG,
            ],
        )
        self.declare_partials(
            Mission.Takeoff.FINAL_MASS,
            Mission.Design.GROSS_MASS,
            val=1.0,
        )
        self.declare_partials(
            Mission.Takeoff.FINAL_MASS,
            Mission.Takeoff.FUEL_SIMPLE,
            val=-1.0,
        )

    def compute(self, inputs, outputs):

        rho_SL = RHO_SEA_LEVEL_METRIC

        v_stall = inputs["v_stall"]
        gross_mass = inputs[Mission.Design.GROSS_MASS]
        ramp_weight = gross_mass * GRAV_ENGLISH_LBM
        rho = inputs["rho"]
        S = inputs[Aircraft.Wing.AREA]
        Cl_max = inputs[Mission.Takeoff.LIFT_COEFFICIENT_MAX]
        thrust = inputs[Mission.Design.THRUST_TAKEOFF_PER_ENG]
        L_over_D = inputs[Mission.Takeoff.LIFT_OVER_DRAG]
        num_engines = self.options['num_engines']
        rho_ratio = rho / rho_SL

        # note: this is different from the paper, not entirely clear why other than rho
        # itself can be eliminated
        rolling_distance = (
            17.0
            * ramp_weight
            / (
                S
                * Cl_max
                * (
                    thrust * num_engines / ramp_weight
                    - (0.20 + 0.00550 * ramp_weight / S) / L_over_D
                )
            )
        )

        rotation_distance = 140.0 * (ramp_weight / (S * Cl_max * rho_ratio)) ** 0.5

        # The calculation below uses all engines operating. The original FLOPS calculation uses one
        # engine inoperative to calculate the field length (not the all-engines performance).
        climbout_distance = (
            140.0
            * (ramp_weight / S) ** 0.5
            / (
                1.0
                + thrust * num_engines / ramp_weight
                - 0.90 / L_over_D
            )
        )

        # The FLOPS methodology does not calculate V2. However, it assumes CL @ V2 = 0.66 * CL_max.
        # Therefore, V2 = V_stall / sqrt(0.66).
        V2 = v_stall * 1.2309

        outputs[Mission.Takeoff.GROUND_DISTANCE] = rolling_distance + \
            rotation_distance + climbout_distance
        outputs[Mission.Takeoff.FINAL_VELOCITY] = V2
        outputs[Mission.Takeoff.FINAL_MASS] = \
            gross_mass - inputs[Mission.Takeoff.FUEL_SIMPLE]
        outputs[Mission.Takeoff.FINAL_ALTITUDE] = 35

    def compute_partials(self, inputs, J):
        rho_SL = RHO_SEA_LEVEL_METRIC

        ramp_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        rho = inputs["rho"]
        S = inputs[Aircraft.Wing.AREA]
        Cl_max = inputs[Mission.Takeoff.LIFT_COEFFICIENT_MAX]
        thrust = inputs[Mission.Design.THRUST_TAKEOFF_PER_ENG]
        L_over_D = inputs[Mission.Takeoff.LIFT_OVER_DRAG]
        num_engines = self.options['num_engines']
        rho_ratio = rho / rho_SL

        den_RD = (
            S
            * Cl_max
            * (
                thrust * num_engines / ramp_weight
                - (0.20 + 0.00550 * ramp_weight / S) / L_over_D
            )
        )
        rad_Rot = ramp_weight / (S * Cl_max * rho_ratio)
        den_Cout = (
            1.0 + thrust * num_engines / ramp_weight - 0.90 / L_over_D
        )

        S * Cl_max * (
            thrust * num_engines / ramp_weight
            - (0.20 + 0.00550 * ramp_weight / S) / L_over_D
        )
        S * Cl_max * (
            -thrust * num_engines / ramp_weight ** 2
            - (0.00550 / S) / L_over_D
        )

        # dRD_dWt = (den_RD*17 - 17*ramp_weight*S*Cl_max*(-thrust*num_engines/ramp_weight**2 - (.00550/S)/L_over_D))/den_RD**2
        dRD_dM = (17 / den_RD - 17 * ramp_weight / den_RD ** 2 * S * Cl_max * (
            -thrust * num_engines / ramp_weight ** 2
            - (0.00550 / S) / L_over_D
        )) * GRAV_ENGLISH_LBM
        dRD_dS = (
            den_RD * 0
            - 17
            * ramp_weight
            * (
                Cl_max
                * (
                    thrust * num_engines / ramp_weight
                    - (0.2 + 0.0055 * ramp_weight / S) / L_over_D
                )
                + S * Cl_max * (0.0055 * ramp_weight / S ** 2 / L_over_D)
            )
        ) / den_RD ** 2
        dRD_dClMax = (
            -17
            * ramp_weight
            / den_RD ** 2
            * (
                S
                * (
                    thrust * num_engines / ramp_weight
                    - (0.20 + 0.00550 * ramp_weight / S) / L_over_D
                )
            )
        )
        dRD_dThrust = (
            -17
            * ramp_weight
            / (
                S
                * Cl_max
                * (
                    thrust * num_engines / ramp_weight
                    - (0.2 + 0.0055 * ramp_weight / S) / L_over_D
                )
            )
            ** 2
            * S
            * Cl_max
            * num_engines
            / ramp_weight
        )
        dRD_dLoD = (
            -17
            * ramp_weight
            / (
                S
                * Cl_max
                * (
                    thrust * num_engines / ramp_weight
                    - (0.2 + 0.0055 * ramp_weight / S) / L_over_D
                )
            )
            ** 2
            * S
            * Cl_max
            * (0.2 + 0.0055 * ramp_weight / S)
            / L_over_D ** 2
        )
        dRD_dRho = 0

        dRot_dM = 140 * 0.5 * \
            rad_Rot ** (-0.5) / (S * Cl_max * rho_ratio) * GRAV_ENGLISH_LBM
        dRot_dS = (
            140
            * 0.5
            * rad_Rot ** (-0.5)
            * (-ramp_weight / (S ** 2 * Cl_max * rho_ratio))
        )
        dRot_dClMax = (
            140
            * 0.5
            * rad_Rot ** (-0.5)
            * (-ramp_weight / (S * Cl_max ** 2 * rho_ratio))
        )
        dRot_dThrust = 0
        dRot_dLoD = 0
        dRot_dRho = (
            140
            * 0.5
            * rad_Rot ** (-0.5)
            * (-ramp_weight / (S * Cl_max * rho ** 2 / rho_SL))
        )

        dCout_dM = (140 * 0.5 * (ramp_weight / S) ** (-0.5) / S / den_Cout - 140 * (
            ramp_weight / S
        ) ** 0.5 / (den_Cout) ** 2 * (
            -thrust * num_engines / ramp_weight ** 2
        )) * GRAV_ENGLISH_LBM
        dCout_dS = (
            140 * 0.5 * (ramp_weight / S) ** (-0.5) * (-ramp_weight) / S ** 2 / den_Cout
        )
        dCout_dClMax = 0
        dCout_dThrust = (
            -140
            * (ramp_weight / S) ** 0.5
            / den_Cout ** 2
            * num_engines
            / ramp_weight
        )
        dCout_dLoD = (
            -140 * (ramp_weight / S) ** 0.5 / den_Cout ** 2 * 0.9 / L_over_D ** 2
        )
        dCout_dRho = 0

        J[Mission.Takeoff.GROUND_DISTANCE,
            Mission.Design.GROSS_MASS] = dRD_dM + dRot_dM + dCout_dM
        J[Mission.Takeoff.GROUND_DISTANCE, "rho"] = \
            dRD_dRho + dRot_dRho + dCout_dRho
        J[Mission.Takeoff.GROUND_DISTANCE,
            Aircraft.Wing.AREA] = dRD_dS + dRot_dS + dCout_dS
        J[
            Mission.Takeoff.GROUND_DISTANCE,
            Mission.Takeoff.LIFT_COEFFICIENT_MAX
        ] = dRD_dClMax + dRot_dClMax + dCout_dClMax
        J[
            Mission.Takeoff.GROUND_DISTANCE,
            Mission.Design.THRUST_TAKEOFF_PER_ENG
        ] = dRD_dThrust + dRot_dThrust + dCout_dThrust
        J[Mission.Takeoff.GROUND_DISTANCE,
            Mission.Takeoff.LIFT_OVER_DRAG] = dRD_dLoD + dRot_dLoD + dCout_dLoD


class TakeoffGroup(om.Group):
    def initialize(self):
        self.options.declare("num_engines", desc="number of engines on aircraft")

    def setup(self):

        self.add_subsystem(
            "USatm",
            USatm1976Comp(num_nodes=1),
            promotes_inputs=[("h", Dynamic.Mission.ALTITUDE)],
            promotes_outputs=["rho", ("sos", "speed_of_sound")],
        )

        self.add_subsystem(
            "stall_speed",
            StallSpeed(),
            promotes_outputs=[
                "v_stall",
            ],
            promotes_inputs=[
                ("mass", Mission.Design.GROSS_MASS),
                "rho",
                ('planform_area', Aircraft.Wing.AREA),
                ("Cl_max", Mission.Takeoff.LIFT_COEFFICIENT_MAX),
            ],
        )

        self.add_subsystem(
            "final_conditions",
            FinalTakeoffConditions(num_engines=self.options["num_engines"]),
            promotes_inputs=[
                "v_stall",
                Mission.Design.GROSS_MASS,
                "rho",
                Aircraft.Wing.AREA,
                Mission.Takeoff.FUEL_SIMPLE,
                Mission.Takeoff.LIFT_COEFFICIENT_MAX,
                Mission.Design.THRUST_TAKEOFF_PER_ENG,
                Mission.Takeoff.LIFT_OVER_DRAG,
            ],
            promotes_outputs=[
                Mission.Takeoff.GROUND_DISTANCE,
                Mission.Takeoff.FINAL_VELOCITY,
                Mission.Takeoff.FINAL_MASS,
                Mission.Takeoff.FINAL_ALTITUDE,
            ],
        )

        self.add_subsystem('compute_mach', om.ExecComp('final_mach = final_velocity / speed_of_sound',
                                                       final_mach={'units': 'unitless'},
                                                       final_velocity={'units': 'm/s'},
                                                       speed_of_sound={'units': 'm/s'}),
                           promotes_inputs=[('speed_of_sound', 'speed_of_sound'),
                                            ('final_velocity', Mission.Takeoff.FINAL_VELOCITY)],
                           promotes_outputs=[('final_mach', Mission.Takeoff.FINAL_MACH)])
